#! /usr/bin/env python

import logging
import warnings

from ..pep8 import old_function, old_method
from ..py2 import (
    dict_keys,
    dict_values,
    force_text,
    is_string,
    ul)

from .parser import (
    XMLFatalError,
    XMLParser)
from .structures import (
    Document,
    DuplicateXMLNAME,
    Element,
    escape_cdsect,
    escape_char_data,
    is_name_char,
    is_name_start_char,
    is_s,
    Node,
    XMLDTD)


#: URI string constant for the special XML namespace
XML_NAMESPACE = "http://www.w3.org/XML/1998/namespace"

#: URI string constant for the special XMLNS namespace
XMLNS_NAMESPACE = "http://www.w3.org/2000/xmlns/"

#: Special string constant used to represent no namespace
NO_NAMESPACE = "~"


class XMLNSError(XMLFatalError):

    """Raised when an illegal QName is found."""
    pass


@old_function('IsValidNCName')
def is_valid_ncname(name):
    """Checks a string against NCName"""
    if name:
        if not is_name_start_char(name[0]) or name[0] == ":":
            return False
        for c in name[1:]:
            if not is_name_char(c) or c == ":":
                return False
        return True
    else:
        return False


def attr_name_key(aname):
    """A nasty function to make sorting attribute names predictable."""
    if is_string(aname):
        return (NO_NAMESPACE, force_text(aname))
    else:
        return aname


class NSNode(Node):

    """Base class for NSElement and Document shared attributes.

    This class adds a number of method for managing the mapping between
    namespace prefixes and namespace URIs in both elements and in the
    document itself.

    You don't have to worry about using these, they are called
    automatically enabling the transparent serialisation of XML elements
    with appropriately defined namespace prefixes.  You only need to use
    these method if you wish to customise the way the mapping is done.
    The most likely use case is simply to call :meth:`make_prefix` at
    the document level to add an explicit declaration of any auxiliary
    namespaces, typically done by the __init__ method on classes derived
    from NSDocument."""

    def __init__(self, parent=None):
        self._prefix_to_ns = {}
        self._ns_to_prefix = {}
        super(NSNode, self).__init__(parent)

    def reset_prefix_map(self, recursive=False):
        self._prefix_to_ns = {}
        self._ns_to_prefix = {}
        if recursive:
            for child in self.get_children():
                if not is_string(child):
                    child.reset_prefix_map(True)

    @old_method('GetPrefix')
    def get_prefix(self, ns):
        """Returns the prefix assigned to a namespace

        ns
            The namespace URI as a character string.

        Returns None if no prefix is currently in force for this
        namespace."""
        if ns == XML_NAMESPACE:
            return 'xml'
        elif ns is None:
            # Attributes with no namespace
            logging.error("Deprecation warning: None for ns")
            import traceback
            traceback.print_stack()
            return ''
        elif ns == NO_NAMESPACE:
            return ''
        prefix = None
        ei = self
        while prefix is None and ei is not None:
            prefix = ei._ns_to_prefix.get(ns, None)
            if prefix is not None:
                # this is the prefix to use, unless it has been reused...
                ej = self
                while ej is not ei:
                    if ej._prefix_to_ns.get(prefix, None) is not None:
                        # so prefix has been reused, keep searching
                        prefix = None
                        break
                    ej = ej.parent
            ei = ei.parent
        return prefix

    def get_ns(self, prefix=''):
        """Returns the namespace associated with prefix.

        prefix
            The prefix to search for, the empty string denotes the
            default namespace.

        This method searches back through the hierarchy until it finds
        the namespace in force or returns None if no definition for this
        prefix can be found.

        In the special case of prefix being 'xml' the XML namespace
        itself is returned.  See :data:`XML_NAMESPACE`."""
        if prefix == 'xml':
            return XML_NAMESPACE
        ns = None
        ei = self
        while ei is not None:
            ns = ei._prefix_to_ns.get(prefix, None)
            if ns:
                break
            if prefix == '':
                ns = getattr(ei, 'default_ns', None)
                if ns:
                    break
            ei = ei.parent
        return ns

    def new_prefix(self, stem='ns'):
        """Return an unused prefix

        stem
            The returned value will be of the form stem# where
            # is a number used in sequence starting with 1.
            This argument defaults to ns so, by default, the prefixes
            ns1, ns2, ns3, etc. are used."""
        ns = 1
        prefix = ''
        while True:
            prefix = "%s%s" % (stem, ns)
            if self.get_ns(prefix) is not None:
                ns = ns + 1
            else:
                break
        return prefix

    @old_method('MakePrefix')
    def make_prefix(self, ns, prefix=None):
        """Creates a new mapping for a namespace

        ns
            The namespace being mapped

        prefix
            The character string representing the namespace in qualified
            names (without the colon).  This parameter is optional, if
            no value is provided then a new randomly generated prefix is
            used using :meth:`new_prefix`.

            Note that an empty string denotes the default namespace,
            which will appear simply as xmlns=<ns> in the element's
            tag.

        If the prefix has already been declared for this node then
        ValueError is raised."""
        if prefix is None:
            prefix = self.new_prefix()
        if prefix in self._prefix_to_ns:
            raise ValueError
        self._prefix_to_ns[prefix] = ns
        self._ns_to_prefix[ns] = prefix
        return prefix

    def get_prefix_map(self):
        """Returns the complete prefix to ns mapping in force

        Combines the prefix mapping for this element with that of it's
        parents.  Returns a dictionary mapping prefix strings to the
        URIs of the namespaces they represent."""
        prefix_map = {}
        ei = self
        while ei is not None:
            prefix_list = dict_keys(ei._prefix_to_ns)
            for prefix in prefix_list:
                if prefix not in prefix_map:
                    prefix_map[prefix] = ei._prefix_to_ns[prefix]
            ei = ei.parent
        return prefix_map

    def write_nsattrs(self, attributes, escape_function=escape_char_data,
                      root=False, **kws):
        """Adds strings representing any namespace attributes

        See :meth:`~pyslet.xml.structures.Element.write_xml_attributes`
        for details of argument usage.

        This method is defined for both NSDocument and NSElement and it
        prefixes the attribute list with any XML namespace declarations
        that are defined by this node.  If *root* is True then all
        namespace declarations that are in force are written, not just
        those attached to this node.  See :meth:`get_prefix_map` for more
        information."""
        escape_function = kws.get('escapeFunction', escape_function)
        ns_attrs = []
        if root:
            prefix_map = self.get_prefix_map()
        else:
            prefix_map = self._prefix_to_ns
        prefix_list = sorted(dict_keys(prefix_map))
        for prefix in prefix_list:
            if prefix:
                ns_attrs.append(
                    ul('xmlns:%s=%s') %
                    (prefix, escape_function(prefix_map[prefix], True)))
            else:
                ns_attrs.append(
                    ul('xmlns=%s') % escape_function(prefix_map[prefix], True))
        attributes[0:0] = ns_attrs


_xmlns_base = (XML_NAMESPACE, 'base')
_xmlns_lang = (XML_NAMESPACE, 'lang')
_xmlns_space = (XML_NAMESPACE, 'space')


class NSElement(NSNode, Element):

    """Element class used for namespace-aware elements.

    Namespace aware elements have special handling for elements that
    contain namespace declarations and for handling qualified names. A
    qualified name is a name that starts with a namespace prefix
    followed by a colon, for example "md:name" might represent the
    'name' element in a particular namespace indicated by the prefix
    'md'.

    The *same* element could equally be encountered with a different prefix
    depending on the namespace declarations in the document.  As a result,
    to interpret element (and attribute) names they must be expanded.

    An expanded name is represented as a 2-tuple consisting of two
    character strings, the first is a URI of a namespace (used only as
    an identifier, the URI does not have to be the URI of an actual
    resource).  The second item is the element name defined within
    that namespace.

    In general, when dealing with classes derived from NSElement you
    should use expanded names wherever you would normally use a plain
    character string.  For example, the class attribute XMLNAME, used by
    derived classes to indicate the default name to use for the element
    the class represents must be an expanded name::

        class MyElement(NSElement):
            XMLNAME = ('http://www.example.com/namespace', 'MyElement')

    Custom attribute mappings use special class attributes with names
    starting with XMLATTR\_ and this mechanism cannot be extended to use
    the expanded names.  As a result these mappings can only be used for
    attributes that have no namespace.  In practice this is not a
    significant limitation as attributes are usually defined this way in
    XML documents.  Note that the special XML attributes (which appear
    to be in the namespace implicitly decared by the prefix "xml:")
    should be referenced using the special purpose get/set methods
    provided."""

    def __init__(self, parent, name=None):
        super(NSElement, self).__init__(parent)
        if name is not None:
            warnings.warn(
                "NSElement: passing name to constructor is deprecated (%s); "
                "use set_xmlname instead" % name)
            import traceback
            traceback.print_stack()
            self.set_xmlname(name)

    def set_xmlname(self, name):
        """Sets the name of this element

        Overridden to support setting the name from either an expanded
        name or an *unqualified* name (in which case the namespace is
        set to None)."""
        if is_string(name):
            self.ns = None
            self.xmlname = name
        elif name is None:
            self.ns = self.xmlname = None
        else:
            self.ns, self.xmlname = name

    def get_xmlname(self):
        """Returns the name of this element

        For classes derived from NSElement this is always an expanded
        name (even if the first component is None, indicating that the
        namespace is not known."""
        return (self.ns, self.xmlname)

    @classmethod
    def mangle_aname(cls, name):
        """Returns a mangled attribute name

        Custom setters are enabled only for attributes with no
        namespace.  For attriubtes from other namespaces the default
        processing defined by the Element's
        set_attribute/get_attribute(s) implementation is used."""
        ns, aname = name
        if ns is None:
            logging.error("Deprecation warning: None for ns")
            import traceback
            traceback.print_stack()
            return "XMLATTR_" + aname
        elif ns == NO_NAMESPACE:
            return "XMLATTR_" + aname
        else:
            return None

    @classmethod
    def unmangle_aname(cls, mname):
        """Overridden to return an expanded name.

        Custom attribute mappings are only supported for attributes with
        no namespace."""
        if mname.startswith('XMLATTR_'):
            return (NO_NAMESPACE, mname[8:])
        else:
            return None

    def set_attribute(self, name, value):
        """Sets the value of an attribute.

        Overridden to allow attributes to be set using either expanded
        names (2-tuples) or *unqualified* names (character strings).

        Implementation notes: for elements descended from NSElement all
        attributes are stored using expanded names internally.  The
        method :meth:`unmangle_name` is overridden to return a 2-tuple
        to make their 'no namespace' designation explicit.

        This method also catches the new namespace prefix mapping for
        the element which is placed in a special attribute by
        :py:meth:`XMLNSParser.parse_nsattrs` and updates the element's
        namespace mappings accordingly."""
        if name == (NO_NAMESPACE, ".ns"):
            self._prefix_to_ns = nsMap = value
            self._ns_to_prefix = dict(
                (k, v) for (k, v) in
                zip(dict_values(nsMap), dict_keys(nsMap)))
            return
        elif (hasattr(self.__class__, 'ID') and
                (name == (NO_NAMESPACE, self.__class__.ID) or
                 name == self.__class__.ID)):
            # we have to override ID handling as mangling is special
            self.set_id(value)
        if is_string(name):
            return Element.set_attribute(self, (NO_NAMESPACE, name), value)
        else:
            return Element.set_attribute(self, name, value)

    def get_attribute(self, name):
        """Gets the value of an attribute.

        Overridden to allow attributes to be got using either expanded
        names (2-tuples) or *unqualified* names (character strings)."""
        if (hasattr(self.__class__, 'ID') and
                (name == (NO_NAMESPACE, self.__class__.ID) or
                 name == self.__class__.ID)):
            # we have to override ID handling as mangling is special
            return self.id
        if is_string(name):
            return Element.get_attribute(self, (NO_NAMESPACE, name))
        else:
            return Element.get_attribute(self, name)

    def is_valid_name(self, value):
        return is_valid_ncname(value)

    @staticmethod
    def sort_names(name_list):
        name_list.sort(key=attr_name_key)

    def get_base(self):
        return self._attrs.get(_xmlns_base, None)

    def set_base(self, base):
        if base is None:
            self._attrs.pop(_xmlns_base, None)
        else:
            self._attrs[_xmlns_base] = str(base)

    def get_lang(self):
        return self._attrs.get(_xmlns_lang, None)

    def set_lang(self, lang):
        if lang is None:
            self._attrs.pop(_xmlns_lang, None)
        else:
            self._attrs[_xmlns_lang] = lang

    def get_space(self):
        return self._attrs.get(_xmlns_space, None)

    def set_space(self, space):
        if space is None:
            self._attrs.pop(_xmlns_space, None)
        else:
            self._attrs[_xmlns_space] = space

    def write_xml_attributes(self, attributes,
                             escape_function=escape_char_data, root=False,
                             **kws):
        escape_function = kws.get('escapeFunction', escape_function)
        attrs = self.get_attributes()
        keys = sorted(dict_keys(attrs))
        for a in keys:
            if is_string(a):
                logging.error(
                    "Deprecation warning: found attribute with no namespace "
                    "in NSElement, %s(%s)", self.__class__.__name__, a)
                aname = a
                prefix = ''
            else:
                ns, aname = a
                prefix = self.get_prefix(ns)
            if prefix is None:
                prefix = self.make_prefix(ns)
            if prefix:
                prefix = prefix + ':'
            attributes.append(
                ul('%s%s=%s') % (prefix, aname, escape_function(attrs[a],
                                                                True)))
        self.write_nsattrs(
            attributes, escape_function=escape_char_data, root=root)

    def generate_xml(self, escape_function=escape_char_data, indent='',
                     tab='\t', root=False, **kws):
        escape_function = kws.get('escapeFunction', escape_function)
        if tab:
            ws = '\n' + indent
            indent = indent + tab
        else:
            ws = ''
        if not self.can_pretty_print():
            # inline all children
            indent = ''
            tab = ''
        attributes = []
        if self.ns:
            # look up the element prefix
            prefix = self.get_prefix(self.ns)
            if prefix is None:
                # We need to declare our namespace
                prefix = self.make_prefix(self.ns, '')
        else:
            prefix = ''
        if prefix:
            prefix = prefix + ':'
        self.write_xml_attributes(attributes, escape_function, root=root)
        if attributes:
            attributes[0:0] = ['']
            attributes = ' '.join(attributes)
        else:
            attributes = ''
        children = self.get_canonical_children()
        try:
            child = next(children)
            if is_string(child) and len(child) and is_s(child[0]):
                # First character is WS, so assume pre-formatted.
                indent = tab = ''
            yield ul('%s<%s%s%s>') % (ws, prefix, self.xmlname, attributes)
            if hasattr(self.__class__, 'SGMLCDATA'):
                # When expressed in SGML this element would have type
                # CDATA so put it in a CDSect
                yield escape_cdsect(self.get_value())
            else:
                while True:
                    if is_string(child):
                        # We force encoding of carriage return as these
                        # are subject to removal
                        yield escape_function(child)
                        # if we have character data content skip closing
                        # ws
                        ws = ''
                    else:
                        try:
                            for s in child.generate_xml(
                                    escape_function, indent, tab):
                                yield s
                        except TypeError:
                            logging.error(
                                "Problem with %s: child was %s",
                                self.__class__.__name__, repr(child))
                            raise
                    try:
                        child = next(children)
                    except StopIteration:
                        break
            if not tab:
                # if we weren't tabbing children we need to skip closing
                # white space
                ws = ''
            yield ul('%s</%s%s>') % (ws, prefix, self.xmlname)
        except StopIteration:
            yield ul('%s<%s%s%s/>') % (ws, prefix, self.xmlname, attributes)


# name provided for backwards compatibility
XMLNSElement = NSElement


class NSDocument(Document, NSNode):

    default_ns = None
    """The default namespace for this document class

    A special class attribute used to set the default namespace for
    elements created within the document that are parsed without an
    effective namespace declaration.  Set to None, but typically
    overridden by derived classes."""

    def XMLParser(self, entity):    # noqa
        """Namespace documents use the special :py:class:`XMLNSParser`.
        """
        return XMLNSParser(entity)

    @classmethod
    def get_element_class(cls, name):
        """Returns a class object suitable for representing <name>

        name is a tuple of (namespace, name), this overrides the
        behaviour of Document, in which name is a string.

        The default implementation returns NSElement."""
        return NSElement


#: name provided for backwards compatibility
XMLNSDocument = NSDocument


class XMLNSParser(XMLParser):

    """A special parser for parsing documents that may use namespaces."""

    _nsdoc_class_table = {}

    @classmethod
    def register_nsdoc_class(cls, doc_class, xname):
        """Registers a document class

        Internally XMLNSParser maintains a single table of document
        classes which can be used to identify the correct class to use
        to represent a document based on the expanded name of the root
        element.

        doc_class
            the class object being registered, it must be derived from
            :py:class:`NSDocument`

        xname
            A tuple of (namespace, name) representing the name of the
            root element.  If either (or both) components are None a
            wildcard is registered that will match any corresponding
            value."""
        cls._nsdoc_class_table[xname] = doc_class

    def get_nsdoc_class(self, xname):
        """Returns a doc class object suitable for this root element

        xname
            An expanded name.

        Returns a class object derived from :py:class:`NSDocument`
        suitable for representing a document with a root element with
        the given expanded name.

        This default implementation uses xname to locate a class
        registered with :meth:`register_nsdoc_class`.  If an exact match
        is not found then wildcard matches are tried matching *only* the
        namespace and root element name in turn.

        A wildcard match is stored in the mapping table either as an
        expanded name of the form (<uri string>, None) or (None,
        <element name>).  The former is preferred as it enables a
        document class to be defined that is capable of representing a
        document with any root element from the given namespace (a
        common use case) and is thus always tried first.

        If no document class can be found, :py:class:`NSDocument` is
        returned."""
        if xname[0] is None:
            doc_class = XMLParser._nsdoc_class_table.get(xname, None)
        else:
            doc_class = XMLParser._nsdoc_class_table.get(xname, None)
            if doc_class is None:
                doc_class = XMLParser.DocumentClassTable.get(
                    (xname[0], None), None)
            if doc_class is None:
                doc_class = XMLParser.DocumentClassTable.get(
                    (None, xname[1]), None)
        if doc_class is None:
            doc_class = NSDocument
        return doc_class

    def __init__(self, entity=None):
        XMLParser.__init__(self, entity)

    def expand_qname(self, qname, ns_defs, use_default=True):
        """Expands a QName, returning a (namespace, name) tuple.

        qname
            The qualified name

        ns_defs
            A mapping of prefix to namespace URI used to expand the name

            If *ns_defs* does not contain a suitable namespace
            definition then the context's existing prefix mapping is
            used, then its parent's mapping is used, and so on.

        use_default (defaults to True)
            Whether or not to return the default namespace for an
            unqualified name.

            If *use_default* is False an unqualified name is returned
            with :py:data:`NO_NAMESPACE` as the namespace (this is used when
            expanding attribute names)."""
        context = self.get_context()
        xname = qname.split(':')
        if len(xname) == 1:
            if qname == 'xmlns':
                return (XMLNS_NAMESPACE, '')
            elif use_default:
                ns_uri = ns_defs.get('', None)
                if ns_uri is None and context is not None:
                    ns_uri = context.get_ns('')
                return (ns_uri, qname)
            else:
                return (NO_NAMESPACE, qname)
        elif len(xname) == 2:
            nsprefix, local = xname
            if nsprefix == 'xml':
                return (XML_NAMESPACE, local)
            elif nsprefix == 'xmlns':
                return (XMLNS_NAMESPACE, local)
            else:
                ns_uri = ns_defs.get(nsprefix, None)
                if ns_uri is None and context is not None:
                    ns_uri = context.get_ns(nsprefix)
                return (ns_uri, local)
        else:
            # something wrong with this element
            raise XMLNSError("Illegal QName: %s" % qname)

    def match_xml_name(self, element, qname):
        """Tests if *qname* is a possible name for this element.

        This method is used by the parser to determine if an end tag is the end
        tag of this element."""
        return element.get_xmlname() == self.expand_qname(qname, {}, True)

    def parse_nsattrs(self, attrs):
        """Manages namespace prefix mappings

        Takes a dictionary of attributes as returned by
        :meth:`parse_stag` and finds any namespace prefix mappings
        returning them as a dictionary of prefix:namespace suitable for
        passing to :py:meth:`expand_qname`.  It also removes the
        namespace declarations from attrs and expands the attribute
        names into (ns, name) pairs.

        Implementation note: a special attribute called '.ns' (in no
        namespace) is set to the parsed prefix mapping dictionary
        enabling the prefix mapping to be passed transparently to
        :py:meth:`NSElement.set_attribute` by py:class:`XMLParser`."""
        ns = {}
        for aname in list(dict_keys(attrs)):
            if aname.startswith('xmlns'):
                if len(aname) == 5:
                    # default ns declaration
                    ns[''] = attrs[aname]
                elif aname[5] == ':':
                    # ns prefix declaration
                    ns[aname[6:]] = attrs[aname]
                del attrs[aname]
        for aname in list(dict_keys(attrs)):
            xname = self.expand_qname(aname, ns, False)
            attrs[xname] = attrs[aname]
            del attrs[aname]
        # Finally, we hide the ns object in the list of attributes so we
        # can retrieve it later Note that '.' is not a valid
        # NameStartChar so we will never collide with a real attribute
        attrs[(NO_NAMESPACE, ".ns")] = ns
        return ns

    def get_stag_class(self, qname, attrs=None):
        """[40] STag

        Overridden to allow for namespace handling."""
        if self.doc is None:
            if self.dtd is None:
                self.dtd = XMLDTD()
            if self.dtd.name is None:
                self.dtd.name = qname
            elif qname is None:
                # document starts with PCDATA, use name declared in DOCTYPE
                qname = self.dtd.name
        # go through attributes and process namespace declarations
        if attrs and not ((NO_NAMESPACE, ".ns") in attrs):
            # This deserves an explanation.  It is possible that
            # get_stag_class will infer an element in sgml_omittag mode
            # forcing the parser to buffer this qname and its associated
            # attributes after we've done namespace expansion of them.
            # There is a real question over whether or not it is safe to
            # buffer expanded attribute names.  It is conceivable that
            # the omitted tag could have FIXED attributes which alter
            # the namespace prefix map.  In theory, we should go back to
            # the original attribute names but smglOmittag mode is a fix
            # up for conforming SGML-style documents into XML (i.e.,
            # HTML).  As such, ignoring this subtle namespacing issue
            # seems reasonable.
            ns = self.parse_nsattrs(attrs)
        else:
            ns = {}
        if qname:
            xname = self.expand_qname(qname, ns)
        else:
            xname = None
        if self.doc is None:
            # we use the expanded name to find the document class, not the DTD
            doc_class = self.get_nsdoc_class(xname)
            self.doc = doc_class()
        else:
            doc_class = self.doc.__class__
        context = self.get_context()
        if qname and xname[0] is None:
            xname = (doc_class.default_ns, xname[1])
        if self.sgml_omittag:
            if qname:
                stag_class = self.doc.get_element_class(xname)
            else:
                stag_class = str
            element_class = context.get_child_class(stag_class)
            if element_class is not stag_class:
                return element_class, None, True
            elif element_class is str:
                # unhanded data, end tag omission not supported
                self.validity_error(
                    "data not allowed in %s (and end-tag omission not "
                    "supported)" % context.__class__.__name__)
                return None, xname, False
            else:
                return element_class, xname, False
        else:
            stag_class = context.get_element_class(xname)
            if stag_class is None:
                stag_class = self.doc.get_element_class(xname)
            return stag_class, xname, False
            # return
            # self.doc.get_element_class(xname),xname,False


@old_function('MapClassElements')
def map_class_elements(class_map, scope, ns_alias_table=None):
    """Adds element name -> class mappings to class_map

    class_map
        A dictionary that maps XML element *expanded* names onto class
        objects that should be used to represent them.

    scope
        A dictionary, or an object containing a __dict__ attribute, that
        will be scanned for class objects to add to the mapping.  This
        enables scope to be a module.  The search is not recursive, to
        add class elements from imported modules you must call
        map_class_elements for each module.

    ns_alias_table
        Used to create multiple mappings for selected element classes
        based on namespace aliases.  It is a dictionary mapping a
        canonical namespace to a list of aliases.  For example, if::

            ns_alias_table={'http://www.example.com/schema-v3': [
                                'http://www.example.com/schema-v2',
                                'http://www.example.com/schema-v1']}

        An element class with::

            XMLNAME = ('http://www.example.com/schema-v3', 'data')

        would then be used by the parser to represent the <data> element
        in the v1, v2 and v3 schema variants.

    The scope is searched for classes derived from :py:class:`NSElement`
    that have an XMLNAME attribute defined.  It is an error if a class
    is found with an XMLNAME that has already been mapped."""
    if not isinstance(scope, dict):
        scope = scope.__dict__
    for name in dict_keys(scope):
        obj = scope[name]
        if issubclass(type(obj), type) and issubclass(obj, NSElement):
            if hasattr(obj, 'XMLNAME'):
                if obj.XMLNAME in class_map:
                    raise DuplicateXMLNAME(
                        "%s and %s have matching XMLNAMEs" %
                        (obj.__name__, class_map[obj.XMLNAME].__name__))
                class_map[obj.XMLNAME] = obj
                if ns_alias_table:
                    aliases = ns_alias_table.get(obj.XMLNAME[0], ())
                    for alias in aliases:
                        alias_name = (alias, obj.XMLNAME[1])
                        if alias_name in class_map:
                            raise DuplicateXMLNAME(
                                "%s and %s have matching XMLNAME alias %s" %
                                (obj.__name__,
                                 class_map[obj.XMLNAME].__name__, alias_name))
                        class_map[alias_name] = obj


@old_function('NSEqualNames')
def match_expanded_names(xname, xmatch, ns_aliases=None):
    """Compares two expanded names

    xname, xmatch
        Expanded names, i.e., 2-tuples of character strings containing
        (namespace URI, element name).

    ns_aliases
        Used to match multiple names based on namespace aliases.  It is
        a list of namespaces that should be treated as equivalent to the
        namespace of xname.  For example::

            match_expanded_names(
                ('http://www.example.com/schema-v3','data'),
                ('http://www.example.com/schema-v1','data'),
                ['http://www.example.com/schema-v2',
                 'http://www.example.com/schema-v1'])

        returns True as xmatch uses an allowed alias for the namespace
        of xname."""
    if xname == xmatch:
        return True
    else:
        for alias in ns_aliases:
            if (alias, xname[1]) == xmatch:
                return True
    return False


@old_function('RegisterNSDocumentClass')
def register_nsdoc_class(doc_class, expanded_name):
    XMLNSParser.register_nsdoc_class(doc_class, expanded_name)



#! /usr/bin/env python

from ..py2 import (
    byte,
    byte_to_bstr,
    byte_value,
    dict_keys,
    is_byte,
    is_unicode,
    join_bytes)
from ..unicode5 import BasicParser, ParserMixin


def is_octet(b):
    """Returns True if a byte matches the production for OCTET."""
    return b is not None and byte_value(b) < 256


def is_char(b):
    """Returns True if a byte matches the production for CHAR."""
    return b is not None and byte_value(b) < 128


def is_upalpha(b):
    """Returns True if a byte matches the production for UPALPHA."""
    return b is not None and 65 <= byte_value(b) <= 90


def is_loalpha(b):
    """Returns True if a byte matches the production for LOALPHA."""
    return b is not None and 97 <= byte_value(b) <= 122


def is_alpha(b):
    """Returns True if a byte matches the production for ALPHA."""
    return is_upalpha(b) or is_loalpha(b)


def is_digit(b):
    """Returns True if a byte matches the production for DIGIT."""
    return b is not None and 48 <= byte_value(b) <= 57


def is_digits(src):
    """Returns True if all bytes match the production for DIGIT.

    Empty strings return False"""
    if src:
        for b in src:
            if not is_digit(b):
                return False
        return True
    else:
        return False


def is_ctl(b):
    """Returns True if a byte matches the production for CTL."""
    return b is not None and (byte_value(b) < 32 or byte_value(b) == 127)


#: octet 13 (carriage return)
CR = byte(13)

#: octet 10 (linefeed)
LF = byte(10)

#: octet 32 (space)
SP = byte(32)

#: octet 9 (horizontal tab)
HT = byte(9)

# : octet 34 (double quote)
DQUOTE = byte(34)

#: the string consisting of CR followed by LF
CRLF = join_bytes([CR, LF])


def is_hex(b):
    """Returns True if a byte matches the production for HEX."""
    return (b is not None and is_digit(b) or (65 <= byte_value(b) <= 70) or
            (97 <= byte_value(b) <= 102))


def is_hexdigits(src):
    """Returns True if all bytes match the production for HEX.

    Empty strings return False"""
    if src:
        for c in src:
            if not is_hex(c):
                return False
        return True
    else:
        return False


SEPARATORS = set('()<>@,;:\\"/[]?={} \t'.encode('ascii'))

LEFT_PARENTHESIS = byte('(')        #: Left Parenthesis "("
RIGHT_PARENTHESIS = byte(')')       #: Right Parentheses "("
LESSTHAN_SIGN = byte('<')           #: Less-than Sign "<"
GREATERTHAN_SIGN = byte('>')        #: Greater-than Sign ">"
COMMERCIAL_AT = byte('@')           #: Commercial At "@"
COMMA = byte(',')                   #: Comma ","
SEMICOLON = byte(';')               #: Semicolon ";"
COLON = byte(':')                   #: Colon ":"
REVERSE_SOLIDUS = byte(0x5C)        #: Reverse solidus "\"
SOLIDUS = byte('/')                 #: Solidus "/"
LEFT_SQUARE_BRACKET = byte('[')     #: Left square bracket "["
RIGHT_SQUARE_BRACKET = byte(']')    #: Right square bracket "["
QUESTION_MARK = byte('?')           #: Question mark "?"
EQUALS_SIGN = byte('=')             #: Equals sign "="
LEFT_CURLY_BRACKET = byte('{')      #: Left curly bracket "{"
RIGHT_CURLY_BRACKET = byte('}')     #: Right curly bracket "}"


def is_separator(b):
    """Returns True if a byte is a separator"""
    return b in SEPARATORS


def check_token(t):
    """Raises ValueError if *t* is *not* a valid token

    t
        A binary string, will also accept a single byte.

    Returns a *character* string representing the token on success."""
    if not isinstance(t, bytes):
        # single byte
        t = bytes((t, ))
    for b in t:
        if b in SEPARATORS:
            raise ValueError("Separator found in token: %s" % t)
        elif is_ctl(b) or not is_char(b):
            raise ValueError("Non-ASCII or CTL found in token: %s" % t)
    return t.decode('iso-8859-1')


def decode_quoted_string(qstring):
    """Decodes a quoted string, returning the unencoded string.

    Surrounding double quotes are removed and quoted bytes, bytes
    preceded by $5C (backslash), are unescaped.

    The return value is a binary string.  In most cases you will want to
    decode it using the latin-1 (iso-8859-1) codec as that was the
    original intention of RFC2616 but in practice anything outside US
    ASCII is likely to be non-portable."""
    qstring = qstring[1:-1]
    if qstring.find(b'\\') >= 0:
        # skip the loop if we don't have any escape sequences
        qbuff = []
        escape = False
        for b in qstring:
            if not escape and b == REVERSE_SOLIDUS:
                escape = True
                continue
            qbuff.append(b)
            escape = False
        return join_bytes(qbuff)
    else:
        return qstring


def quote_string(s, force=True):
    """Places a string in double quotes, returning the quoted string.

    force
        Always quote the string, defaults to True.  If False then valid
        tokens are not quoted but returned as-is.

    This is the reverse of :py:func:`decode_quoted_string`.  Note that
    only the double quote, \\ and CTL characters other than SP and HT
    are quoted in the output."""
    qstring = [DQUOTE]
    for b in s:
        if b in (REVERSE_SOLIDUS, DQUOTE) or (is_ctl(b) and b not in (SP, HT)):
            qstring.append(REVERSE_SOLIDUS)
            qstring.append(b)
            force = True
        elif is_ctl(b) or is_separator(b):
            force = True
            qstring.append(b)
        else:
            qstring.append(b)
    if force:
        qstring.append(DQUOTE)
    else:
        del qstring[0]
    return join_bytes(qstring)


def format_parameters(parameters):
    """Formats a dictionary of parameters

    This function is suitable for formatting parameter dictionaries
    parsed by :py:meth:`WordParser.parse_parameters`.  These
    dictionaries are key/value pairs where the keys are *character*
    strings and the values are *binary* strings.

    Parameter values are quoted only if their values require it, that
    is, only if their values are *not* valid tokens."""
    keys = sorted(dict_keys(parameters))
    format = []
    for k in keys:
        format.append(b'; ')
        p, v = parameters[k]
        format.append(p.encode('ascii'))
        format.append(b'=')
        format.append(quote_string(v, force=False))
    return b''.join(format)


class OctetParser(BasicParser):

    """A special purpose parser for parsing HTTP productions

    Strictly speaking, HTTP operates only on bytes so the parser is
    always set to binary mode.  However, as a concession to the various
    normative references to HTTP in other specifications where character
    strings are parsed they will be accepted provided they only contain
    *US ASCII* characters."""

    def __init__(self, source):
        if is_unicode(source):
            source = source.encode('ascii')
        super(OctetParser, self).__init__(source)

    def parse_lws(self):
        """Parses a single instance of the production LWS

        The return value is the LWS string parsed or None if there is no
        LWS."""
        savepos = self.pos
        lws = []
        if self.parse(CRLF):
            lws += [CR, LF]
        splen = 0
        while True:
            b = self.parse_one((SP, HT))
            if b is None:
                break
            else:
                lws.append(b)
                splen += 1
        if lws and splen:
            return join_bytes(lws)
        self.setpos(savepos)
        return None

    def parse_onetext(self, unfold=False):
        """Parses a single TEXT instance.

        unfold
            Pass True to replace folding LWS with a single SP. Defaults
            to False.

        Parses a single byte or run of LWS matching the production TEXT.
        The return value is either:

            1   a single byte of TEXT (*not* a binary string) excluding
                the LWS characters

            2   a binary string of LWS

            3   None if no TEXT was found

        You may find the utility function :func:`pyslet.py2.is_byte`
        useful to distinguish cases 1 and 2 correctly in both Python 2
        and Python 3."""
        lws = self.parse_lws()
        if lws:
            if unfold and lws[:2] == CRLF:
                return byte_to_bstr(SP)
            else:
                return lws
        elif is_octet(self.the_char) and not is_ctl(self.the_char):
            result = self.the_char
            self.next_char()
            return result
        else:
            return None

    def parse_text(self, unfold=False):
        """Parses TEXT

        unfold
            Pass True to replace folding LWS with a single SP. Defaults
            to False.

        Parses a run of characters matching the production TEXT.  The
        return value is the matching TEXT as a binary string (including
        any LWS) or None if no TEXT was found."""
        text = []
        while self.the_char is not None:
            t = self.parse_onetext(unfold)
            if t is not None:
                if is_byte(t):
                    text.append(t)
                else:
                    text += list(t)
            else:
                break
        if text:
            return join_bytes(text)
        else:
            return None

    def parse_token(self):
        """Parses a token.

        Parses a single instance of the production token.  The return
        value is the matching token as a binary string or None if no
        token was found."""
        token = []
        while self.the_char is not None:
            if is_ctl(self.the_char) or is_separator(self.the_char):
                break
            else:
                token.append(self.the_char)
                self.next_char()
        if token:
            return join_bytes(token)
        else:
            return None

    def parse_comment(self, unfold=False):
        """Parses a comment.

        unfold
            Pass True to replace folding LWS with a single SP. Defaults
            to False.

        Parses a single instance of the production comment.  The return
        value is the entire matching comment as a binary string
        (including the brackets, quoted pairs and any nested comments)
        or None if no comment was found."""
        if not self.parse(b"("):
            return None
        comment = [b"("]
        depth = 1
        while self.the_char is not None:
            if self.parse(b")"):
                comment.append(b")")
                depth -= 1
                if depth < 1:
                    break
            elif self.parse(b"("):
                comment.append(b"(")
                depth += 1
            elif self.match(b"\\"):
                qp = self.parse_quoted_pair()
                if qp:
                    comment.append(qp)
                else:
                    self.parser_error("quoted pair")
            else:
                ctext = self.parse_ctext(unfold)
                if ctext:
                    comment.append(ctext)
                else:
                    break
        if depth:
            self.parser_error("comment close")
        else:
            return b''.join(comment)

    def parse_ctext(self, unfold=False):
        """Parses ctext.

        unfold
            Pass True to replace folding LWS with a single SP. Defaults
            to False.

        Parses a run of characters matching the production ctext.  The
        return value is the matching ctext as a binary string (including
        any LWS) or None if no ctext was found.

        The original text of RFC2616 is ambiguous in the definition of
        ctext but the later errata_ corrected this to exclude the
        backslash byte ($5C) so we stop if we encounter one.

        .. _errata: https://www.rfc-editor.org/errata_search.php?rfc=2616
        """
        ctext = []
        while self.the_char is not None:
            if self.match_one(b"()\\"):
                break
            else:
                t = self.parse_onetext(unfold)
                if t is not None:
                    if is_byte(t):
                        ctext.append(t)
                    else:
                        ctext += list(t)
                else:
                    break
        if ctext:
            return join_bytes(ctext)
        else:
            return None

    def parse_quoted_string(self, unfold=False):
        """Parses a quoted-string.

        unfold
            Pass True to replace folding LWS with a single SP. Defaults
            to False.

        Parses a single instance of the production quoted-string.  The
        return value is the entire matching string (including the quotes
        and any quoted pairs) or None if no quoted-string was found."""
        if not self.parse(b'"'):
            return None
        qs = [b'"']
        while self.the_char is not None:
            if self.parse(b'"'):
                qs.append(b'"')
                break
            elif self.match(b"\\"):
                qp = self.parse_quoted_pair()
                if qp:
                    qs.append(qp)
                else:
                    self.parser_error("quoted pair")
            else:
                qdtext = self.parse_qdtext(unfold)
                if qdtext:
                    qs.append(qdtext)
                else:
                    self.parser_error('<">')
        return b''.join(qs)

    def parse_qdtext(self, unfold=False):
        """Parses qdtext.

        Parses a run of characters matching the production qdtext.  The
        return value is the matching qdtext string (including any LWS)
        or None if no qdtext was found.

        If *unfold* is True then any folding LWS is replaced with a
        single SP.  It defaults to False

        Although the production for qdtext would include the backslash
        character we stop if we encounter one, following the RFC2616
        errata_ instead."""
        qdtext = []
        while self.the_char is not None:
            if self.match_one(b'"\\'):
                break
            else:
                t = self.parse_onetext(unfold)
                if t is not None:
                    if is_byte(t):
                        qdtext.append(t)
                    else:
                        qdtext += list(t)
                else:
                    break
        if qdtext:
            return join_bytes(qdtext)
        else:
            return None

    def parse_quoted_pair(self):
        """Parses a single quoted-pair.

        The return value is the matching binary string including the
        backslash so it will always be of length 2 or None if no
        quoted-pair was found."""
        savepos = self.pos
        if self.parse(b"\\"):
            if is_char(self.the_char):
                qdpair = b"\\" + byte_to_bstr(self.the_char)
                self.next_char()
                return qdpair
            else:
                self.setpos(savepos)
        return None


class BadSyntax(ValueError):

    """Raised by the :class:`WordParser`

    Whenever a syntax error is encountered by the parsers.  Note that
    tokenization errors are raised separately during construction itself.

    production
        The name of the production being parsed.  (Defaults to an empty
        string.)

    parser
        The :class:`WordParser` instance raising the error (optional)

    BadSyntax is a subclass of ValueError."""

    def __init__(self, production='', parser=None):
        self.production = production
        if parser:
            #: the position of the parser when the error was raised
            if parser.match_end():
                self.pos = len(parser.source)
            else:
                self.pos = parser.src_pos[parser.pos]
            #: up to 40 characters/bytes to the left of pos
            self.left = parser.source[max(0, self.pos - 40):self.pos]
            #: up to 40 characters/bytes to the right of pos
            self.right = parser.source[self.pos:self.pos + 40]
            if production:
                msg = "SyntaxError: expected %s at [%i]" % (production,
                                                            self.pos)
            else:
                msg = "SyntaxError: at [%i]" % self.pos
        else:
            self.pos = None
            self.left = None
            self.right = None
            if production:
                msg = "SyntaxError: expected %s" % production
            else:
                msg = "SyntaxError"
        ValueError.__init__(self, msg)


class WordParser(ParserMixin):

    """A word-level parser and tokeniser for the HTTP grammar.

    source
        The binary string to be parsed into words.  It will normally be
        valid TEXT but it can contain control characters if they are
        escaped as part of a comment or quoted string.  For
        compatibility, character strings are accepted provided they only
        contain US ASCII characters

    ingore_sp (defaults to True)
        LWS is unfolded automatically.  By default the parser ignores
        spaces according to the rules for implied LWS in the
        specification and neither SP nor HT will be stored in the word
        list.  If you set *ignore_sp* to False then LWS is not ignored
        and each run of LWS is returned as a single SP in the word list.

    The source is parsed completely into words on construction using
    :class:`OctetParser`. If the source contains a CRLF (or any other
    non-TEXT bytes) that is not part of a folding or escape sequence it
    raises :class:`~pyslet.unicode5.ParserError`.

    For the purposes of this parser, a word may be either a single byte
    (in which case it is a separator or SP, note that HT is never stored
    in the word list) or a binary string, in which case it is a token, a
    comment or a quoted string.  Warning: in Python 2 a single byte is
    indistinguishable from a binary string of length 1.

    Methods follow the same pattern as that described in the related
    :class:`pyslet.unicode5.BasicParser` using match\_, parse\_ and
    require\_ naming conventions.  It also includes the
    :class:`pyslet.unicode5.ParseMixin` class to enable the convenience
    methods for converting between look-ahead and non-look-ahead parsing
    modes."""

    def __init__(self, source, ignore_sp=True):
        self.source = source
        self.words = []
        self.src_pos = []
        #: a pointer to the current word in the list
        self.pos = 0
        #: the current word or None
        self.the_word = None
        self.last_error = None
        self._init_parser(source, ignore_sp)

    def _init_parser(self, source, ignore_sp=True):
        p = OctetParser(source)
        while p.the_char is not None:
            sp = False
            src_pos = p.pos
            while p.parse_lws():
                if not ignore_sp:
                    sp = True
            if sp:
                self.words.append(SP)
                self.src_pos.append(src_pos)
                src_pos = p.pos
            if is_separator(p.the_char):
                if p.the_char == LEFT_PARENTHESIS:
                    self.words.append(p.parse_comment(True))
                elif p.the_char == DQUOTE:
                    self.words.append(p.parse_quoted_string(True))
                else:
                    self.words.append(p.the_char)
                    p.next_char()
                self.src_pos.append(src_pos)
            elif p.the_char is None:
                break
            else:
                self.words.append(p.require_production(p.parse_token(),
                                                       "TEXT"))
                self.src_pos.append(src_pos)
        self.pos = 0
        if self.words:
            self.the_word = self.words[0]
        else:
            self.the_word = None

    def setpos(self, pos):
        """Sets the current position of the parser.

        Example usage for look-ahead::

                # wp is a WordParser instance
                savepos=wp.pos
                try:
                        # parse a token/sub-token combination
                        token=wp.require_token()
                        wp.require_separator(byte('/'))
                        subtoken=wp.require_token()
                        return token,subtoken
                except BadSyntax:
                        wp.setpos(savepos)
                        return None,None"""
        self.pos = pos
        if self.pos < len(self.words):
            self.the_word = self.words[self.pos]
        else:
            self.the_word = None

    def parser_error(self, production=None):
        """Raises an error encountered by the parser

        See :class:`BadSyntax` for details.

        If production is None then the previous error is re-raised. If
        multiple errors have been raised previously the one with the
        most advanced parser position is used.  The operation is similar
        to :meth:`pyslet.unicode5.BasicParser.parser_error`.

        To improve the quality of error messages an internal record of
        the starting position of each word is kept (within the original
        source).

        The position of the parser is always set to the position of the
        error raised."""
        if production:
            e = BadSyntax(production, self)
        elif self.last_error is not None and self.pos <= self.last_error.pos:
            e = self.last_error
        else:
            e = BadSyntax('', self)
        if self.last_error is None or e.pos > self.last_error.pos:
            self.last_error = e
        if e.pos != self.pos:
            self.setpos(e.pos)
        raise e

    def match_end(self):
        """True if all of :attr:`words` have been parsed"""
        return self.the_word is None

    def peek(self):
        """Returns the next word

        If there are no more words, returns None."""
        if self.the_word:
            return self.the_word
        else:
            return ""

    def parse_word(self):
        """Parses any word from the list

        Returns the word parsed or None if the parser was already at the
        end of the word list."""
        result = self.the_word
        self.pos += 1
        if self.pos < len(self.words):
            self.the_word = self.words[self.pos]
        else:
            self.the_word = None
        return result

    def parse_word_as_bstr(self):
        """Parses any word from the list

        Returns a binary string representing the word.  In cases where
        the next work is a separator it converts the word to a binary
        string (in Python 2 this is a noop) before returning it."""
        w = self.parse_word()
        if w is not None and not isinstance(w, bytes):
            w = byte_to_bstr(w)
        return w

    def is_token(self):
        """Returns True if the current word is a token"""
        # words are never empty!
        return isinstance(self.the_word, bytes) and \
            self.the_word[0] not in SEPARATORS

    def parse_token(self):
        """Parses a token from the list of words

        Returns the token or None if the next word was not a token.  The
        return value is a binary string.  This is consistent with the
        use of this method for parsing tokens in contexts where a token
        or a quoted string may be present."""
        if self.is_token():
            return self.parse_word()
        else:
            return None

    def parse_tokenlower(self):
        """Returns a lower-cased token parsed from the word list

        Returns None if the next word was not a token.  Unlike
        :meth:`parse_token` the result is a *character* string."""
        if self.is_token():
            return self.parse_word().lower().decode('ascii')
        else:
            return None

    def parse_tokenlist(self):
        """Parses a list of tokens

        Returns the list or [] if no tokens were found.  Lists are
        defined by RFC2616 as being comma-separated.  Note that empty
        items are ignored, so strings such as "x,,y" return just ["x",
        "y"].

        The list of tokens is returned as a list of character strings."""
        result = []
        while self.the_word:
            token = self.parse_token()
            if token:
                result.append(token.decode('ascii'))
            elif self.parse_separator(COMMA):
                continue
            else:
                break
        return result

    def require_token(self, expected="token"):
        """Returns the current token or raises :py:class:`BadSyntax`

        expected
            the name of the expected production, it defaults to
            "token"."""
        token = self.parse_token()
        if token is None:
            self.parser_error(expected)
        else:
            return token

    def is_integer(self):
        """Returns True if the current word is an integer token"""
        return self.the_word and is_digits(self.the_word)

    def parse_integer(self):
        """Parses an integer token from the list of words

        Return the integer's *value* or None."""
        if self.is_integer():
            return int(self.parse_word())
        else:
            return None

    def require_integer(self, expected="integer"):
        """Parses an integer or raises :py:class:`BadSyntax`

        expected
            can be set to the name of the expected object, defaults to
            "integer"."""
        result = self.parse_integer()
        if result is None:
            self.parser_error(expected)
        else:
            return result

    def is_hexinteger(self):
        """Returns True if the current word is a hex token"""
        return self.the_word and is_hexdigits(self.the_word)

    def parse_hexinteger(self):
        """Parses a hex integer token from the list of words

        Return the hex integer's *value* or None."""
        if self.is_hexinteger():
            return int(self.parse_word(), 16)
        else:
            return None

    def require_hexinteger(self, expected="hex integer"):
        """Parses a hex integer or raises :py:class:`BadSyntax`

        expected
            can be set to the name of the expected object, defaults to
            "hex integer"."""
        result = self.parse_hexinteger()
        if result is None:
            self.parser_error(expected)
        else:
            return result

    def is_separator(self, sep):
        """Returns True if the current word matches *sep*"""
        return self.the_word == sep

    def parse_separator(self, sep):
        """Parses a *sep* from the list of words.

        Returns True if the current word matches *sep* and False
        otherwise."""
        if self.the_word == sep:
            self.parse_word()
            return True
        else:
            return False

    def require_separator(self, sep, expected=None):
        """Parses *sep* or raises :py:class:`BadSyntax`

        sep
            A separtor byte (not a binary string).

        expected
            can be set to the name of the expected object"""
        if self.the_word == sep:
            self.parse_word()
        else:
            self.parser_error(expected)

    def is_quoted_string(self):
        """Returns True if the current word is a quoted string."""
        return isinstance(self.the_word, bytes) and self.the_word[0] == DQUOTE

    def parse_quoted_string(self):
        """Parses a quoted string from the list of words.

        Returns the *decoded* value of the quoted string or None."""
        if self.is_quoted_string():
            return decode_quoted_string(self.parse_word())
        else:
            return None

    def parse_sp(self):
        """Parses a SP from the list of words.

        Returns True if the current word is a SP and False otherwise."""
        return self.parse_separator(SP)

    def parse_parameters(self, parameters, ignore_allsp=True,
                         case_sensitive=False, qmode=None):
        """Parses a set of parameters

        parameters
            the dictionary in which to store the parsed parameters

        ignore_allsp
            a boolean (defaults to True) which causes the function to
            ignore all LWS in the word list.  If set to False then space
            around the '=' separator is treated as an error and raises
            BadSyntax.

        case_sensitive
            controls whether parameter names are treated as case
            sensitive, defaults to False.

        qmode
            allows you to pass a special parameter name that will
            terminate parameter parsing (without being parsed itself).
            This is used to support headers such as the "Accept" header
            in which the parameter called "q" marks the boundary between
            media-type parameters and Accept extension parameters.
            Defaults to None

        Updates the parameters dictionary with the new parameter
        definitions. The key in the dictionary is the parameter name
        (converted to lower case if parameters are being dealt with case
        insensitively) and the value is a 2-item tuple of (name, value)
        always preserving the original case of the parameter name.

        Returns the parameters dictionary as the result.  The method
        always succeeds as parameter lists can be empty.

        Compatibility warning: parameter names must be tokens and are
        therefore converted to *character* strings.  Parameter values,
        on the other hand, may be quoted strings containing characters
        from unknown character sets and are therefore always represented
        as binary strings."""
        self.parse_sp()
        while self.the_word:
            savepos = self.pos
            try:
                self.parse_sp()
                self.require_separator(SEMICOLON)
                self.parse_sp()
                param_name = self.require_token(
                    "parameter").decode('iso-8859-1')
                if not case_sensitive:
                    param_key = param_name.lower()
                else:
                    param_key = param_name
                if qmode and param_key == qmode:
                    raise BadSyntax
                if ignore_allsp:
                    self.parse_sp()
                self.require_separator(EQUALS_SIGN, "parameter")
                if ignore_allsp:
                    self.parse_sp()
                if self.is_token():
                    param_value = self.parse_token()
                elif self.is_quoted_string():
                    param_value = self.parse_quoted_string()
                else:
                    self.parser_error("parameter value")
                parameters[param_key] = (param_name, param_value)
            except BadSyntax:
                self.setpos(savepos)
                break
        return parameters

    def parse_remainder(self, sep=b''):
        """Parses the rest of the words

        The result is a single string representing the remaining words
        joined with *sep*, which defaults to an *empty* string.

        Returns an empty string if the parser is at the end of the word
        list."""
        if self.the_word:
            result = sep.join(w if isinstance(w, bytes) else byte_to_bstr(w)
                              for w in self.words[self.pos:])
        else:
            result = b''
        self.setpos(len(self.words))
        return result



#! /usr/bin/env python
"""This module implements test to check against the IMSCC Profile 1.1
specification defined by IMS GLC"""

from . import imscc_profilev1p0 as v1p0
from . import imscpv1p2 as imscp
from .xml import namespace as xmlns
from .xml import xsdatatypes as xsi


IMSCC_CP_NAMESPACE = "http://www.imsglobal.org/xsd/imsccv1p1/imscp_v1p1"
IMSCC_CP_SCHEMALOCATION = ("http://www.imsglobal.org/profile/cc/ccv1p2/"
                           "ccv1p2_imscp_v1p2_v1p0.xsd")

IMSCC_LOMMANIFEST_NAMESPACE = "http://ltsc.ieee.org/xsd/imsccv1p2/LOM/manifest"
IMSCC_LOMMANIFEST_SCHEMALOCATION = ("http://www.imsglobal.org/profile/cc/"
                                    "ccv1p2/LOM/ccv1p2_lommanifest_v1p0.xsd")

IMSCC_LOMRESOURCE_NAMESPACE = "http://ltsc.ieee.org/xsd/imsccv1p2/LOM/resource"
IMSCC_LOMRESOURCE_SCHEMALOCATION = ("http://www.imsglobal.org/profile/cc/"
                                    "ccv1p2/LOM/ccv1p2_lomresource_v1p0.xsd")


class Schema(imscp.Schema):
    XMLNAME = (IMSCC_CP_NAMESPACE, 'schema')


class SchemaVersion(imscp.SchemaVersion):
    XMLNAME = (IMSCC_CP_NAMESPACE, 'schemaversion')


class Metadata(imscp.Metadata):
    XMLNAME = (IMSCC_CP_NAMESPACE, 'metadata')
    SchemaClass = Schema
    SchemaVersionClass = SchemaVersion


class Organization(imscp.Organization):
    XMLNAME = (IMSCC_CP_NAMESPACE, 'organization')


class Organizations(imscp.Organizations):
    XMLNAME = (IMSCC_CP_NAMESPACE, 'organizations')
    OrganizationClass = Organization


class File(imscp.File):
    XMLNAME = (IMSCC_CP_NAMESPACE, 'file')


class Dependency(imscp.Dependency):
    XMLNAME = (IMSCC_CP_NAMESPACE, 'dependency')


class Resource(imscp.Resource):
    XMLNAME = (IMSCC_CP_NAMESPACE, 'resource')
    MetadataClass = Metadata
    FileClass = File
    DependencyClass = Dependency


class Resources(imscp.Resources):
    XMLNAME = (IMSCC_CP_NAMESPACE, 'resources')
    ResourceClass = Resource


class Manifest(imscp.Manifest):
    XMLNAME = (IMSCC_CP_NAMESPACE, 'manifest')
    MetadataClass = Metadata
    OrganizationsClass = Organizations
    ResourcesClass = Resources
    ManifestClass = None

Manifest.ManifestClass = Manifest


class ManifestDocument(xmlns.XMLNSDocument):
    classMap = {}

    def __init__(self, **args):
        """"""
        xmlns.XMLNSDocument.__init__(self, **args)
        self.defaultNS = IMSCC_CP_NAMESPACE
        self.make_prefix(xsi.XMLSCHEMA_NAMESPACE, 'xsi')
        self.make_prefix(IMSCC_LOMMANIFEST_NAMESPACE, 'lomimscc')
        self.make_prefix(IMSCC_LOMRESOURCE_NAMESPACE, 'lom')
        schema_location = [IMSCC_CP_NAMESPACE, IMSCC_CP_SCHEMALOCATION,
                           IMSCC_LOMMANIFEST_NAMESPACE,
                           IMSCC_LOMMANIFEST_SCHEMALOCATION,
                           IMSCC_LOMRESOURCE_NAMESPACE,
                           IMSCC_LOMRESOURCE_SCHEMALOCATION]
        if isinstance(self.root, imscp.CPElement):
            self.root.set_attribute(
                (xsi.XMLSCHEMA_NAMESPACE, 'schemaLocation'),
                ' '.join(schema_location))

    def get_element_class(self, name):
        eclass = ManifestDocument.classMap.get(
            name, ManifestDocument.classMap.get((name[0], None),
                                                xmlns.XMLNSElement))
        return eclass

xmlns.map_class_elements(ManifestDocument.classMap, globals())
# xmlns.map_class_elements(ManifestDocument.classMap,imsmd)
# xmlns.map_class_elements(ManifestDocument.classMap,imsqti)
# Add other supported metadata schemas in here


class ContentPackage(imscp.ContentPackage):
    ManifestDocumentClass = ManifestDocument


class CommonCartridge(v1p0.CommonCartridge):
    ContentPackageClass = ContentPackage
